

/* Copyright 2022 the SumatraPDF project authors (see AUTHORS file).
   License: Simplified BSD (see COPYING.BSD) */

#include "utils/BaseUtil.h"
#include "utils/VecSegmented.h"
#include "utils/ScopedWin.h"
#include "utils/WinUtil.h"

#include "wingui/UIModels.h"

#include "wingui/Layout.h"
#include "wingui/Window.h"

#include "wingui/wingui2.h"

#include "utils/Log.h"

// TODO: call RemoveWindowSubclass in WM_NCDESTROY as per
// https://devblogs.microsoft.com/oldnewthing/20031111-00/?p=41883

#define DEFAULT_WIN_CLASS L"WC_WIN32_WINDOW"

static UINT_PTR g_subclassId = 0;

UINT_PTR NextSubclassId() {
    g_subclassId++;
    return g_subclassId;
}

// initial value which should be save
static int g_currCtrlID = 100;

int GetNextCtrlID() {
    ++g_currCtrlID;
    return g_currCtrlID;
}

// a way to register for messages for a given hwnd / msg combo

struct HwndMsgHandler {
    HWND hwnd;
    UINT msg;
    void* user = nullptr;
    void (*handler)(void* user, WndEvent* ev);
};

VecSegmented<HwndMsgHandler> gHwndMsgHandlers;

void WindowCleanup() {
    gHwndMsgHandlers.allocator.FreeAll();
}

static void ClearHwndMsgHandler(HwndMsgHandler* h) {
    CrashIf(!h);
    if (!h) {
        return;
    }
    h->hwnd = nullptr;
    h->msg = 0;
    h->user = nullptr;
    h->handler = nullptr;
}

static HwndMsgHandler* FindHandlerForHwndAndMsg(HWND hwnd, UINT msg, bool create) {
    CrashIf(hwnd == nullptr);
    for (auto h : gHwndMsgHandlers) {
        if (h->hwnd == hwnd && h->msg == msg) {
            return h;
        }
    }
    if (!create) {
        return nullptr;
    }
    // we might have free slot
    HwndMsgHandler* res = nullptr;
    for (auto h : gHwndMsgHandlers) {
        if (h->hwnd == nullptr) {
            res = h;
            break;
        }
    }
    if (!res) {
        res = gHwndMsgHandlers.AllocAtEnd();
    }
    ClearHwndMsgHandler(res);
    res->hwnd = hwnd;
    res->msg = msg;
    return res;
}

void RegisterHandlerForMessage(HWND hwnd, UINT msg, void (*handler)(void* user, WndEvent*), void* user) {
    auto h = FindHandlerForHwndAndMsg(hwnd, msg, true);
    h->handler = handler;
    h->user = user;
}

void UnregisterHandlerForMessage(HWND hwnd, UINT msg) {
    auto h = FindHandlerForHwndAndMsg(hwnd, msg, false);
    ClearHwndMsgHandler(h);
}

static void UnregisterHandlersForHwnd(HWND hwnd) {
    for (auto h : gHwndMsgHandlers) {
        if (h->hwnd == hwnd) {
            ClearHwndMsgHandler(h);
        }
    }
}

// TODO: potentially more messages
// https://docs.microsoft.com/en-us/cpp/mfc/reflected-window-message-ids?view=vs-2019
static HWND GetChildHWNDForMessage(UINT msg, WPARAM wp, LPARAM lp) {
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-ctlcolorbtn
    if (WM_CTLCOLORBTN == msg) {
        return (HWND)lp;
    }
    if (WM_CTLCOLORSTATIC == msg) {
        HDC hdc = (HDC)wp;
        return WindowFromDC(hdc);
    }
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-notify
    if (WM_NOTIFY == msg) {
        NMHDR* hdr = (NMHDR*)lp;
        return hdr->hwndFrom;
    }
    // https://docs.microsoft.com/en-us/windows/win32/menurc/wm-command
    if (WM_COMMAND == msg) {
        return (HWND)lp;
    }
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-drawitem
    if (WM_DRAWITEM == msg) {
        DRAWITEMSTRUCT* s = (DRAWITEMSTRUCT*)lp;
        return s->hwndItem;
    }
    // https://docs.microsoft.com/en-us/windows/win32/menurc/wm-contextmenu
    if (WM_CONTEXTMENU == msg) {
        return (HWND)wp;
    }
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-vscroll--trackbar-
    if (WM_VSCROLL == msg) {
        return (HWND)lp;
    }
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-hscroll--trackbar-
    if (WM_HSCROLL == msg) {
        return (HWND)lp;
    }

    // TODO: there's no HWND so have to do it differently e.g. allocate
    // unique CtlID, store it in WindowBase and compare that
#if 0
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-measureitem
    if (WM_MEASUREITEM == msg) {
        MEASUREITEMSTRUCT* s = (MEASUREITEMSTRUCT*)lp;
        return s->CtlID;
    }
#endif
    return nullptr;
}

bool HandleRegisteredMessages(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp, LRESULT& res) {
    HWND hwndLookup = hwnd;
    HWND hwndMaybe = GetChildHWNDForMessage(msg, wp, lp);
    if (hwndMaybe != nullptr) {
        hwndLookup = hwndMaybe;
    }
    auto h = FindHandlerForHwndAndMsg(hwndLookup, msg, false);
    if (!h) {
        return false;
    }
    WndEvent ev;
    SetWndEventSimple(ev);
    h->handler(h->user, &ev);
    res = ev.result;
    return ev.didHandle;
}

// to ensure we never overflow control ids
// we reset the counter in Window::Window(),
// because ids only need to be unique within window
// this works as long as we don't interleave creation
// of windows and controls in those windows
void ResetCtrlID() {
    g_currCtrlID = 100;
}

// http://www.guyswithtowels.com/blog/10-things-i-hate-about-win32.html#ModelessDialogs
// to implement a standard dialog navigation we need to call
// IsDialogMessage(hwnd) in message loop.
// hwnd has to be current top-level window that is modeless dialog
// we need to manually maintain this window
HWND g_currentModelessDialog = nullptr;

HWND GetCurrentModelessDialog() {
    return g_currentModelessDialog;
}

// set to nullptr to disable
void SetCurrentModelessDialog(HWND hwnd) {
    g_currentModelessDialog = hwnd;
}

CopyWndEvent::CopyWndEvent(WndEvent* dst, WndEvent* src) {
    this->dst = dst;
    this->src = src;
    dst->hwnd = src->hwnd;
    dst->msg = src->msg;
    dst->lp = src->lp;
    dst->wp = src->wp;
    dst->w = src->w;
}

CopyWndEvent::~CopyWndEvent() {
    src->didHandle = dst->didHandle;
    src->result = dst->result;
}

Kind kindWindowBase = "windowBase";

static LRESULT wndBaseProcDispatch(WindowBase* w, HWND hwnd, UINT msg, WPARAM wp, LPARAM lp, bool& didHandle) {
    CrashIf(hwnd != w->hwnd);

    // or maybe get rid of WindowBase::WndProc and use msgFilterInternal
    // when per-control custom processing is needed
    if (w->msgFilter) {
        WndEvent ev{};
        SetWndEvent(ev);
        w->msgFilter(&ev);
        if (ev.didHandle) {
            didHandle = true;
            return ev.result;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-ctlcolorbtn
    if (WM_CTLCOLORBTN == msg) {
        auto bgBrush = w->backgroundColorBrush;
        if (bgBrush != nullptr) {
            didHandle = true;
            return (LRESULT)bgBrush;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-ctlcolorstatic
    if (WM_CTLCOLORSTATIC == msg) {
        HDC hdc = (HDC)wp;
        if (w->textColor != ColorUnset) {
            SetTextColor(hdc, w->textColor);
            // SetTextColor(hdc, RGB(255, 255, 255));
            didHandle = true;
        }
        auto bgBrush = w->backgroundColorBrush;
        if (bgBrush != nullptr) {
            SetBkMode(hdc, TRANSPARENT);
            didHandle = true;
            return (LRESULT)bgBrush;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/winmsg/wm-size
    if (WM_SIZE == msg) {
        if (!w->onSize) {
            return 0;
        }
        SizeEvent ev;
        SetWndEvent(ev);
        ev.dx = LOWORD(lp);
        ev.dy = HIWORD(lp);
        w->onSize(&ev);
        if (ev.didHandle) {
            didHandle = true;
            return 0;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/menurc/wm-command
    if (WM_COMMAND == msg) {
        if (!w->onWmCommand) {
            return 0;
        }
        WmCommandEvent ev{};
        SetWndEvent(ev);
        ev.id = LOWORD(wp);
        ev.ev = HIWORD(wp);
        w->onWmCommand(&ev);
        if (ev.didHandle) {
            didHandle = true;
            return ev.result;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-keydown
    // https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-keyup
    if ((WM_KEYUP == msg) || (WM_KEYDOWN == msg)) {
        if (!w->onKeyDownUp) {
            return 0;
        }
        KeyEvent ev{};
        SetWndEvent(ev);
        ev.isDown = (WM_KEYDOWN == msg);
        ev.keyVirtCode = (int)wp;
        w->onKeyDownUp(&ev);
        if (ev.didHandle) {
            didHandle = true;
            // 0 means: did handle
            return 0;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-char
    if (WM_CHAR == msg) {
        if (!w->onChar) {
            return 0;
        }
        CharEvent ev{};
        SetWndEvent(ev);
        ev.keyCode = (int)wp;
        w->onChar(&ev);
        if (ev.didHandle) {
            didHandle = true;
            // 0 means: did handle
            return 0;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-mousewheel
    if (msg == WM_MOUSEWHEEL || msg == WM_MOUSEHWHEEL) {
        if (!w->onMouseWheel) {
            return 0;
        }
        MouseWheelEvent ev{};
        SetWndEvent(ev);
        ev.isVertical = (msg == WM_MOUSEWHEEL);
        ev.delta = GET_WHEEL_DELTA_WPARAM(wp);
        ev.keys = GET_KEYSTATE_WPARAM(wp);
        ev.x = GET_X_LPARAM(lp);
        ev.y = GET_Y_LPARAM(lp);
        w->onMouseWheel(&ev);
        if (ev.didHandle) {
            didHandle = true;
            return 0;
        }
    }

    // https://docs.microsoft.com/en-us/windows/win32/shell/wm-dropfiles
    if (msg == WM_DROPFILES) {
        if (!w->onDropFiles) {
            return 0;
        }

        DropFilesEvent ev{};
        SetWndEvent(ev);
        ev.hdrop = (HDROP)wp;
        // TODO: docs say it's always zero but sumatra code elsewhere
        // treats 0 and 1 differently
        CrashIf(lp != 0);
        w->onDropFiles(&ev);
        if (ev.didHandle) {
            didHandle = true;
            return 0; // 0 means: did handle
        }
    }

    // handle the rest in WndProc
    WndEvent ev{};
    SetWndEvent(ev);
    w->WndProc(&ev);
    didHandle = ev.didHandle;
    return ev.result;
}

static LRESULT CALLBACK wndProcCustom(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp) {
    // auto msgName = GetWinMessageName(msg);
    // logf("hwnd: 0x%6p, msg: 0x%03x (%s), wp: 0x%x\n", hwnd, msg, msgName, wp);

    if (WM_NCCREATE == msg) {
        CREATESTRUCT* cs = (CREATESTRUCT*)lp;
        Window* w = (Window*)cs->lpCreateParams;
        w->hwnd = hwnd;
        SetWindowLongPtrW(hwnd, GWLP_USERDATA, (LONG_PTR)w);
        return DefWindowProc(hwnd, msg, wp, lp);
    }

    LRESULT res = 0;

    res = wg::TryReflectMessages(hwnd, msg, wp, lp);
    if (res != 0) {
        return res;
    }

    if (HandleRegisteredMessages(hwnd, msg, wp, lp, res)) {
        return res;
    }

    // TODDO: a hack, a Window might be deleted when we get here
    // happens e.g. when we call CloseWindow() inside
    // wndproc. Maybe instead of calling DestroyWindow()
    // we should delete WindowInfo, for proper shutdown sequence
    if (WM_DESTROY == msg) {
        return DefWindowProc(hwnd, msg, wp, lp);
    }

    Window* w = (Window*)GetWindowLongPtr(hwnd, GWLP_USERDATA);
    if (!w) {
        return DefWindowProc(hwnd, msg, wp, lp);
    }

    // this is the last message ever received by hwnd
    // TODO: move it to wndBaseProcDispatch? Maybe they don't
    // need WM_*DESTROY notifications?
    if (WM_NCDESTROY == msg) {
        if (w->onDestroy) {
            WindowDestroyEvent ev{};
            SetWndEvent(ev);
            ev.window = w;
            w->onDestroy(&ev);
            return 0;
        }
        return DefWindowProc(hwnd, msg, wp, lp);
    }

    // TODO: should this go into WindowBase?
    if (WM_CLOSE == msg) {
        if (w->onClose) {
            WindowCloseEvent ev{};
            SetWndEvent(ev);
            w->onClose(&ev);
            if (ev.cancel) {
                return 0;
            }
        }
        return DefWindowProc(hwnd, msg, wp, lp);
    }

    if (w->isDialog) {
        // TODO: should handle more messages as per
        // https://stackoverflow.com/questions/35688400/set-full-focus-on-a-button-setfocus-is-not-enough
        // and https://docs.microsoft.com/en-us/windows/win32/dlgbox/dlgbox-programming-considerations
        if (WM_ACTIVATE == msg) {
            if (wp == 0) {
                // becoming inactive
                SetCurrentModelessDialog(nullptr);
            } else {
                // becoming active
                SetCurrentModelessDialog(w->hwnd);
            }
        }
    }

    if (WM_PAINT == msg) {
        PAINTSTRUCT ps;
        BeginPaint(hwnd, &ps);
        auto bgBrush = w->backgroundColorBrush;
        if (bgBrush != nullptr) {
            FillRect(ps.hdc, &ps.rcPaint, bgBrush);
        }
        EndPaint(hwnd, &ps);
        return 0;
    }

    bool didHandle = false;
    res = wndBaseProcDispatch(w, hwnd, msg, wp, lp, didHandle);
    if (didHandle) {
        return res;
    }
    res = DefWindowProcW(hwnd, msg, wp, lp);
    // auto msgName = GetWinMessageName(msg);
    // logf("hwnd: 0x%6p, msg: 0x%03x (%s), wp: 0x%x, res: 0x%x\n", hwnd, msg, msgName, wp, res);
    return res;
}

static LRESULT CALLBACK wndProcSubclassed(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp, UINT_PTR uIdSubclass,
                                          DWORD_PTR dwRefData) {
    CrashIf(dwRefData == 0);
    WindowBase* w = (WindowBase*)dwRefData;

    if (uIdSubclass != w->subclassId) {
        return DefSubclassProc(hwnd, msg, wp, lp);
    }

    LRESULT res = 0;
    if (HandleRegisteredMessages(hwnd, msg, wp, lp, res)) {
        return res;
    }

    bool didHandle = false;
    res = wndBaseProcDispatch(w, hwnd, msg, wp, lp, didHandle);
    if (didHandle) {
        return res;
    }
    return DefSubclassProc(hwnd, msg, wp, lp);
}

// TODO: do I need WM_CTLCOLORSTATIC?
#if 0
    // https://docs.microsoft.com/en-us/windows/win32/controls/wm-ctlcolorstatic
    if (WM_CTLCOLORSTATIC == msg) {
        HDC hdc = (HDC)wp;
        if (w->textColor != ColorUnset) {
            SetTextColor(hdc, w->textColor);
        }
        auto bgBrush = w->backgroundColorBrush;
        if (bgBrush != nullptr) {
            // SetBkMode(hdc, TRANSPARENT);
            return (LRESULT)bgBrush;
        }
    }
#endif

void WindowBase::Subclass() {
    CrashIf(!hwnd);
    WindowBase* wb = this;
    subclassId = NextSubclassId();
    BOOL ok = SetWindowSubclass(hwnd, wndProcSubclassed, subclassId, (DWORD_PTR)wb);
    CrashIf(!ok);
    if (!ok) {
        subclassId = 0;
    }
}

void WindowBase::Unsubclass() {
    if (subclassId) {
        RemoveWindowSubclass(hwnd, wndProcSubclassed, subclassId);
        subclassId = 0;
    }
}

WindowBase::WindowBase() {
    kind = kindWindowBase;
    ctrlID = GetNextCtrlID();
}

// generally not needed for child controls as they are destroyed when
// a parent is destroyed
void WindowBase::Destroy() {
    auto tmp = hwnd;
    if (IsWindow(tmp)) {
        DestroyWindow(tmp);
        tmp = nullptr;
    }
    hwnd = nullptr;
}

WindowBase::~WindowBase() {
    Unsubclass();
    if (backgroundColorBrush != nullptr) {
        DeleteObject(backgroundColorBrush);
    }
    Destroy();
    UnregisterHandlersForHwnd(hwnd);
}

void WindowBase::WndProc(WndEvent* ev) {
    ev->didHandle = false;
}

Size WindowBase::GetIdealSize() {
    return {};
}

void Handle_WM_CONTEXTMENU(WindowBase* w, WndEvent* ev) {
    CrashIf(ev->msg != WM_CONTEXTMENU);
    CrashIf(!w->onContextMenu);
    // https://docs.microsoft.com/en-us/windows/win32/menurc/wm-contextmenu
    ContextMenuEvent cmev;
    CopyWndEvent cpev(&cmev, ev);
    cmev.w = w;
    cmev.mouseGlobal.x = GET_X_LPARAM(ev->lp);
    cmev.mouseGlobal.y = GET_Y_LPARAM(ev->lp);
    POINT pt{cmev.mouseGlobal.x, cmev.mouseGlobal.y};
    if (pt.x != -1) {
        MapWindowPoints(HWND_DESKTOP, w->hwnd, &pt, 1);
    }
    cmev.mouseWindow.x = pt.x;
    cmev.mouseWindow.y = pt.y;
    w->onContextMenu(&cmev);
    ev->didHandle = true;
}

static void Dispatch_WM_CONTEXTMENU(void* user, WndEvent* ev) {
    WindowBase* w = (WindowBase*)user;
    Handle_WM_CONTEXTMENU(w, ev);
}

bool WindowBase::Create(HWND parent) {
    auto h = GetModuleHandle(nullptr);
    int x = CW_USEDEFAULT;
    if (initialPos.x != -1) {
        x = initialPos.x;
    }
    int y = CW_USEDEFAULT;
    if (initialPos.y != -1) {
        y = initialPos.y;
    }

    int dx = CW_USEDEFAULT;
    if (initialSize.dx > 0) {
        dx = initialSize.dx;
    }
    int dy = CW_USEDEFAULT;
    if (initialSize.dy > 0) {
        dy = initialSize.dy;
    }
    HMENU m = (HMENU)(UINT_PTR)ctrlID;
    hwnd = CreateWindowExW(dwExStyle, winClass, L"", dwStyle, x, y, dx, dy, parent, m, h, nullptr);
    CrashIf(!hwnd);

    if (hwnd == nullptr) {
        return false;
    }

    if (onDropFiles != nullptr) {
        DragAcceptFiles(hwnd, TRUE);
    }

    // TODO: maybe always register so that we can set onContextMenu
    // after creation
    if (onContextMenu) {
        void* user = this;
        RegisterHandlerForMessage(hwnd, WM_CONTEXTMENU, Dispatch_WM_CONTEXTMENU, user);
    }

    if (hfont == nullptr) {
        hfont = GetDefaultGuiFont();
    }
    SetFont(hfont);
    HwndSetText(hwnd, text.AsView());
    return true;
}

void WindowBase::SuspendRedraw() const {
    SendMessageW(hwnd, WM_SETREDRAW, FALSE, 0);
}

void WindowBase::ResumeRedraw() const {
    SendMessageW(hwnd, WM_SETREDRAW, TRUE, 0);
}

void WindowBase::SetFocus() const {
    ::SetFocus(hwnd);
}

bool WindowBase::IsFocused() const {
    BOOL isFocused = ::IsFocused(hwnd);
    return tobool(isFocused);
}

void WindowBase::SetIsEnabled(bool isEnabled) const {
    // TODO: make it work even if not yet created?
    CrashIf(!hwnd);
    BOOL enabled = isEnabled ? TRUE : FALSE;
    ::EnableWindow(hwnd, enabled);
}

bool WindowBase::IsEnabled() const {
    BOOL enabled = ::IsWindowEnabled(hwnd);
    return tobool(enabled);
}

Kind WindowBase::GetKind() {
    return kind;
}

void WindowBase::SetVisibility(Visibility newVisibility) {
    // TODO: make it work before Create()?
    CrashIf(!hwnd);
    visibility = newVisibility;
    bool isVisible = IsVisible();
    // TODO: a different way to determine if is top level vs. child window?
    if (GetParent(hwnd) == nullptr) {
        ::ShowWindow(hwnd, isVisible ? SW_SHOW : SW_HIDE);
    } else {
        BOOL bIsVisible = toBOOL(isVisible);
        SetWindowStyle(hwnd, WS_VISIBLE, bIsVisible);
    }
}

Visibility WindowBase::GetVisibility() {
    return visibility;
#if 0
    if (GetParent(hwnd) == nullptr) {
        // TODO: what to do for top-level window?
        CrashMe();
        return true;
    }
    bool isVisible = IsWindowStyleSet(hwnd, WS_VISIBLE);
    return isVisible;
#endif
}

// convenience function
void WindowBase::SetIsVisible(bool isVisible) {
    SetVisibility(isVisible ? Visibility::Visible : Visibility::Collapse);
}

bool WindowBase::IsVisible() const {
    return visibility == Visibility::Visible;
}

void WindowBase::SetFont(HFONT f) {
    hfont = f;
    HwndSetFont(hwnd, f);
}

HFONT WindowBase::GetFont() const {
    HFONT res = hfont;
    if (!res) {
        res = HwndGetFont(hwnd);
    }
    if (!res) {
        res = GetDefaultGuiFont();
    }
    return res;
}

void WindowBase::SetIcon(HICON iconIn) {
    hIcon = iconIn;
    HwndSetIcon(hwnd, hIcon);
}

HICON WindowBase::GetIcon() const {
    return hIcon;
}

void WindowBase::SetText(const WCHAR* s) {
    auto str = ToUtf8Temp(s);
    SetText(str.AsView());
}

void WindowBase::SetText(std::string_view sv) {
    text.Set(sv);
    // can be set before we create the window
    if (hwnd) {
        HwndSetText(hwnd, text.AsView());
        HwndInvalidate(hwnd);
    }
}

std::string_view WindowBase::GetText() {
    auto sw = win::GetTextTemp(hwnd);
    auto sa = ToUtf8Temp(sw.AsView());
    text.Set(sa.AsView());
    return text.AsView();
}

void WindowBase::SetTextColor(COLORREF col) {
    if (ColorNoChange == col) {
        return;
    }
    textColor = col;
    // can be set before we create the window
    if (!hwnd) {
        return;
    }
    InvalidateRect(hwnd, nullptr, FALSE);
}

void WindowBase::SetBackgroundColor(COLORREF col) {
    if (col == ColorNoChange) {
        return;
    }
    backgroundColor = col;
    if (backgroundColorBrush != nullptr) {
        DeleteObject(backgroundColorBrush);
        backgroundColorBrush = nullptr;
    }
    if (backgroundColor != ColorUnset) {
        backgroundColorBrush = CreateSolidBrush(backgroundColor);
    }
    // can be set before we create the window
    if (!hwnd) {
        return;
    }
    InvalidateRect(hwnd, nullptr, FALSE);
}

void WindowBase::SetColors(COLORREF bg, COLORREF txt) {
    SetBackgroundColor(bg);
    SetTextColor(txt);
}

void WindowBase::SetRtl(bool isRtl) const {
    SetWindowExStyle(hwnd, WS_EX_LAYOUTRTL | WS_EX_NOINHERITLAYOUT, isRtl);
}

Kind kindWindow = "window";

static void RegisterWindowClass(Window* w) {
    WNDCLASSEXW wc{};
    BOOL ok = ::GetClassInfoExW(GetInstance(), w->winClass, &wc);
    if (ok) {
        return;
    }
    wc.cbSize = sizeof(wc);
    wc.hIcon = w->hIcon;
    wc.hCursor = LoadCursorW(nullptr, IDC_ARROW);
    wc.hIconSm = w->hIconSm;
    wc.lpfnWndProc = wndProcCustom;
    wc.lpszClassName = w->winClass;
    wc.lpszMenuName = w->lpszMenuName;
    ATOM atom = RegisterClassExW(&wc);
    CrashIf(!atom);
}

Window::Window() {
    ResetCtrlID();
    kind = kindWindow;
    dwExStyle = 0;
    dwStyle = WS_OVERLAPPEDWINDOW;
}

bool Window::Create(HWND parent) {
    if (parent == nullptr) {
        dwStyle |= WS_CLIPCHILDREN;
    } else {
        dwStyle |= WS_CHILD;
    }
    if (winClass == nullptr) {
        winClass = DEFAULT_WIN_CLASS;
    }
    RegisterWindowClass(this);

    int x = CW_USEDEFAULT;
    if (initialPos.x != -1) {
        x = initialPos.x;
    }
    int y = CW_USEDEFAULT;
    if (initialPos.y != -1) {
        y = initialPos.y;
    }

    int dx = CW_USEDEFAULT;
    if (initialSize.dx > 0) {
        dx = initialSize.dx;
    }
    int dy = CW_USEDEFAULT;
    if (initialSize.dy > 0) {
        dy = initialSize.dy;
    }
    auto title = ToWstrTemp(this->text.AsView());
    HINSTANCE hinst = GetInstance();
    hwnd = CreateWindowExW(dwExStyle, winClass, title, dwStyle, x, y, dx, dy, parent, nullptr, hinst, (void*)this);
    CrashIf(!hwnd);
    if (!hwnd) {
        return false;
    }
    if (hfont == nullptr) {
        hfont = GetDefaultGuiFont();
    }
    // trigger creating a backgroundBrush
    SetBackgroundColor(backgroundColor);
    SetFont(hfont);
    SetIcon(hIcon);
    HwndSetText(hwnd, text.AsView());
    return true;
}

Window::~Window() = default;

void Window::SetTitle(std::string_view title) {
    SetText(title);
}

void Window::Close() {
    ::SendMessageW(hwnd, WM_CLOSE, 0, 0);
}

// if only top given, set them all to top
// if top, right given, set bottom to top and left to right
void WindowBase::SetInsetsPt(int top, int right, int bottom, int left) {
    insets = DpiScaledInsets(hwnd, top, right, bottom, left);
}

Size WindowBase::Layout(const Constraints bc) {
    dbglayoutf("WindowBase::Layout() %s ", GetKind());
    LogConstraints(bc, "\n");

    auto hinset = insets.left + insets.right;
    auto vinset = insets.top + insets.bottom;
    auto innerConstraints = bc.Inset(hinset, vinset);

    int dx = MinIntrinsicWidth(0);
    int dy = MinIntrinsicHeight(0);
    childSize = innerConstraints.Constrain(Size{dx, dy});
    auto res = Size{
        childSize.dx + hinset,
        childSize.dy + vinset,
    };
    return res;
}

int WindowBase::MinIntrinsicHeight(int) {
#if 0
    auto vinset = insets.top + insets.bottom;
    Size s = GetIdealSize();
    return s.dy + vinset;
#else
    Size s = GetIdealSize();
    return s.dy;
#endif
}

int WindowBase::MinIntrinsicWidth(int) {
#if 0
    auto hinset = insets.left + insets.right;
    Size s = GetIdealSize();
    return s.dx + hinset;
#else
    Size s = GetIdealSize();
    return s.dx;
#endif
}

void WindowBase::SetPos(RECT* r) const {
    ::MoveWindow(hwnd, r);
}

void WindowBase::SetBounds(Rect bounds) {
    dbglayoutf("WindowBaseLayout:SetBounds() %s %d,%d - %d, %d\n", GetKind(), bounds.x, bounds.y, bounds.dx, bounds.dy);

    lastBounds = bounds;

    bounds.x += insets.left;
    bounds.y += insets.top;
    bounds.dx -= (insets.right + insets.left);
    bounds.dy -= (insets.bottom + insets.top);

    auto r = RectToRECT(bounds);
    ::MoveWindow(hwnd, &r);
    // TODO: optimize if doesn't change position
    ::InvalidateRect(hwnd, nullptr, TRUE);
}

#if 0
void WindowBase::SetBounds(const RECT& r) const {
    SetPos((RECT*)&r);
}
#endif

int RunMessageLoop(HACCEL accelTable, HWND hwndDialog) {
    MSG msg;
    while (GetMessage(&msg, nullptr, 0, 0)) {
        if (wg::PreTranslateMessage(msg)) {
            continue;
        }
        if (TranslateAccelerator(msg.hwnd, accelTable, &msg)) {
            continue;
        }
        if (hwndDialog && IsDialogMessage(hwndDialog, &msg)) {
            continue;
        }
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    return (int)msg.wParam;
}

// TODO: support accelerator table?
// TODO: a better way to stop the loop e.g. via shared
// atomic int to signal termination and sending WM_IDLE
// to trigger processing of the loop
void RunModalWindow(HWND hwndDialog, HWND hwndParent) {
    if (hwndParent != nullptr) {
        EnableWindow(hwndParent, FALSE);
    }

    MSG msg;
    bool isFinished = false;
    while (!isFinished) {
        BOOL ok = WaitMessage();
        if (!ok) {
            DWORD err = GetLastError();
            LogLastError(err);
            isFinished = true;
            continue;
        }
        while (!isFinished && PeekMessage(&msg, nullptr, 0, 0, PM_REMOVE)) {
            if (msg.message == WM_QUIT) {
                isFinished = true;
                break;
            }
            if (!IsDialogMessage(hwndDialog, &msg)) {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
    }

    if (hwndParent != nullptr) {
        EnableWindow(hwndParent, TRUE);
    }
}

// sets initial position of w within hwnd. Assumes w->initialSize is set.
void PositionCloseTo(WindowBase* w, HWND hwnd) {
    CrashIf(!hwnd);
    Size is = w->initialSize;
    CrashIf(is.IsEmpty());
    RECT r{};
    BOOL ok = GetWindowRect(hwnd, &r);
    CrashIf(!ok);

    // position w in the the center of hwnd
    // if window is bigger than hwnd, let the system position
    // we don't want to hide it
    int offX = (RectDx(r) - is.dx) / 2;
    if (offX < 0) {
        return;
    }
    int offY = (RectDy(r) - is.dy) / 2;
    if (offY < 0) {
        return;
    }
    Point& ip = w->initialPos;
    ip.x = (int)r.left + (int)offX;
    ip.y = (int)r.top + (int)offY;
}

// TODDO: add rest of messages:
// https://codeeval.dev/gist/9f8b444a5a181fbb6391d304b2dace52

struct WinMsgWithName {
    UINT msg;
    const char* name;
};

#define dm(n) \
    { n, #n }

// https://social.msdn.microsoft.com/Forums/windowsapps/en-US/f677f319-9f02-4438-92fb-6e776924425d/windowproc-and-messages-0x90-0x91-0x92-0x93?forum=windowsuidevelopment
#define WM_UAHDESTROYWINDOW 0x0090
#define WM_UAHDRAWMENU 0x0091
#define WM_UAHDRAWMENUITEM 0x0092
#define WM_UAHINITMENU 0x0093
#define WM_UAHMEASUREMENUITEM 0x0094
#define WM_UAHNCPAINTMENUPOPUP 0x0095

static WinMsgWithName gWinMessageNames[] = {
    dm(WM_NULL),
    dm(WM_CREATE),
    dm(WM_DESTROY),
    dm(WM_MOVE),
    dm(WM_SIZE),
    dm(WM_ACTIVATE),
    dm(WM_SETFOCUS),
    dm(WM_KILLFOCUS),
    dm(WM_ENABLE),
    dm(WM_SETREDRAW),
    dm(WM_SETTEXT),
    dm(WM_GETTEXT),
    dm(WM_GETTEXTLENGTH),
    dm(WM_PAINT),
    dm(WM_CLOSE),
    dm(WM_QUERYENDSESSION),
    dm(WM_QUERYOPEN),
    dm(WM_ENDSESSION),
    dm(WM_QUIT),
    dm(WM_ERASEBKGND),
    dm(WM_SYSCOLORCHANGE),
    dm(WM_SHOWWINDOW),
    dm(WM_WININICHANGE),
    dm(WM_SETTINGCHANGE),
    dm(WM_DEVMODECHANGE),
    dm(WM_ACTIVATEAPP),
    dm(WM_FONTCHANGE),
    dm(WM_TIMECHANGE),
    dm(WM_CANCELMODE),
    dm(WM_SETCURSOR),
    dm(WM_MOUSEACTIVATE),
    dm(WM_CHILDACTIVATE),
    dm(WM_QUEUESYNC),
    dm(WM_GETMINMAXINFO),
    dm(WM_PAINTICON),
    dm(WM_ICONERASEBKGND),
    dm(WM_NEXTDLGCTL),
    dm(WM_SPOOLERSTATUS),
    dm(WM_DRAWITEM),
    dm(WM_MEASUREITEM),
    dm(WM_DELETEITEM),
    dm(WM_VKEYTOITEM),
    dm(WM_CHARTOITEM),
    dm(WM_SETFONT),
    dm(WM_GETFONT),
    dm(WM_SETHOTKEY),
    dm(WM_GETHOTKEY),
    dm(WM_QUERYDRAGICON),
    dm(WM_COMPAREITEM),
    dm(WM_GETOBJECT),
    dm(WM_COMPACTING),
    dm(WM_COMMNOTIFY),
    dm(WM_WINDOWPOSCHANGING),
    dm(WM_WINDOWPOSCHANGED),
    dm(WM_POWER),
    dm(WM_COPYDATA),
    dm(WM_CANCELJOURNAL),
    dm(WM_NOTIFY),
    dm(WM_INPUTLANGCHANGEREQUEST),
    dm(WM_INPUTLANGCHANGE),
    dm(WM_TCARD),
    dm(WM_HELP),
    dm(WM_USERCHANGED),
    dm(WM_NOTIFYFORMAT),
    dm(WM_CONTEXTMENU),
    dm(WM_STYLECHANGING),
    dm(WM_STYLECHANGED),
    dm(WM_DISPLAYCHANGE),
    dm(WM_GETICON),
    dm(WM_SETICON),
    dm(WM_NCCREATE),
    dm(WM_NCDESTROY),
    dm(WM_NCCALCSIZE),
    dm(WM_NCHITTEST),
    dm(WM_NCPAINT),
    dm(WM_NCACTIVATE),
    dm(WM_GETDLGCODE),
    dm(WM_SYNCPAINT),
    dm(WM_NCMOUSEMOVE),
    dm(WM_NCLBUTTONDOWN),
    dm(WM_NCLBUTTONUP),
    dm(WM_NCLBUTTONDBLCLK),
    dm(WM_NCRBUTTONDOWN),
    dm(WM_NCRBUTTONUP),
    dm(WM_NCRBUTTONDBLCLK),
    dm(WM_NCMBUTTONDOWN),
    dm(WM_NCMBUTTONUP),
    dm(WM_NCMBUTTONDBLCLK),

    dm(EM_GETSEL),
    dm(EM_SETSEL),
    dm(EM_GETRECT),
    dm(EM_SETRECT),
    dm(EM_SETRECTNP),
    dm(EM_SCROLL),
    dm(EM_LINESCROLL),
    dm(EM_SCROLLCARET),
    dm(EM_GETMODIFY),
    dm(EM_SETMODIFY),
    dm(EM_GETLINECOUNT),
    dm(EM_LINEINDEX),
    dm(EM_SETHANDLE),
    dm(EM_GETHANDLE),
    dm(EM_GETTHUMB),
    dm(EM_LINELENGTH),
    dm(EM_REPLACESEL),
    dm(EM_GETLINE),
    dm(EM_LIMITTEXT),
    dm(EM_CANUNDO),
    dm(EM_UNDO),
    dm(EM_FMTLINES),
    dm(EM_LINEFROMCHAR),
    dm(EM_SETTABSTOPS),
    dm(EM_SETPASSWORDCHAR),
    dm(EM_EMPTYUNDOBUFFER),
    dm(EM_GETFIRSTVISIBLELINE),
    dm(EM_SETREADONLY),
    dm(EM_SETWORDBREAKPROC),
    dm(EM_GETWORDBREAKPROC),
    dm(EM_GETPASSWORDCHAR),
    dm(EM_SETMARGINS),
    dm(EM_GETMARGINS),
    dm(EM_GETLIMITTEXT),
    dm(EM_POSFROMCHAR),
    dm(EM_CHARFROMPOS),
    dm(EM_SETIMESTATUS),
    dm(EM_GETIMESTATUS),
    dm(EM_ENABLEFEATURE),

    dm(WM_TOUCH),
    dm(WM_NCPOINTERUPDATE),
    dm(WM_NCPOINTERDOWN),
    dm(WM_NCPOINTERUP),
    dm(WM_POINTERUPDATE),
    dm(WM_POINTERDOWN),
    dm(WM_POINTERUP),
    dm(WM_POINTERENTER),
    dm(WM_POINTERLEAVE),
    dm(WM_POINTERACTIVATE),
    dm(WM_POINTERCAPTURECHANGED),
    dm(WM_TOUCHHITTESTING),
    dm(WM_POINTERWHEEL),
    dm(WM_POINTERHWHEEL),
    dm(DM_POINTERHITTEST),
    dm(WM_POINTERROUTEDTO),
    dm(WM_POINTERROUTEDAWAY),
    dm(WM_POINTERROUTEDRELEASED),

    dm(WM_IME_SETCONTEXT),
    dm(WM_IME_NOTIFY),
    dm(WM_IME_CONTROL),
    dm(WM_IME_COMPOSITIONFULL),
    dm(WM_IME_SELECT),
    dm(WM_IME_CHAR),
    dm(WM_IME_REQUEST),
    dm(WM_IME_KEYDOWN),
    dm(WM_IME_KEYUP),

    dm(WM_MOUSEHOVER),
    dm(WM_MOUSELEAVE),

    dm(WM_MENUSELECT),
    dm(WM_MENUCHAR),
    dm(WM_ENTERIDLE),
    dm(WM_MENURBUTTONUP),
    dm(WM_MENUDRAG),
    dm(WM_MENUGETOBJECT),
    dm(WM_UNINITMENUPOPUP),
    dm(WM_MENUCOMMAND),
    dm(WM_CHANGEUISTATE),
    dm(WM_UPDATEUISTATE),
    dm(WM_QUERYUISTATE),
    dm(WM_CTLCOLORMSGBOX),
    dm(WM_CTLCOLOREDIT),
    dm(WM_CTLCOLORLISTBOX),
    dm(WM_CTLCOLORBTN),
    dm(WM_CTLCOLORDLG),
    dm(WM_CTLCOLORSCROLLBAR),
    dm(WM_CTLCOLORSTATIC),
    dm(MN_GETHMENU),

    dm(WM_MOUSEMOVE),
    dm(WM_LBUTTONDOWN),
    dm(WM_LBUTTONUP),
    dm(WM_LBUTTONDBLCLK),
    dm(WM_RBUTTONDOWN),
    dm(WM_RBUTTONUP),
    dm(WM_RBUTTONDBLCLK),
    dm(WM_MBUTTONDOWN),
    dm(WM_MBUTTONUP),
    dm(WM_MBUTTONDBLCLK),
    dm(WM_MOUSEWHEEL),
    dm(WM_XBUTTONDOWN),
    dm(WM_XBUTTONUP),
    dm(WM_XBUTTONDBLCLK),

    dm(WM_MOUSEHWHEEL),
    dm(WM_NEXTMENU),
    dm(WM_PARENTNOTIFY),
    dm(WM_ENTERMENULOOP),
    dm(WM_EXITMENULOOP),
    dm(WM_SIZING),
    dm(WM_CAPTURECHANGED),
    dm(WM_MOVING),

    dm(WM_NCXBUTTONDOWN),
    dm(WM_NCXBUTTONUP),
    dm(WM_NCXBUTTONDBLCLK),
    dm(WM_INPUT_DEVICE_CHANGE),
    dm(WM_INPUT),
    dm(WM_KEYFIRST),
    dm(WM_KEYDOWN),
    dm(WM_KEYUP),
    dm(WM_CHAR),
    dm(WM_DEADCHAR),
    dm(WM_SYSKEYDOWN),
    dm(WM_SYSKEYUP),
    dm(WM_SYSCHAR),
    dm(WM_SYSDEADCHAR),
    dm(WM_UNICHAR),
    dm(WM_KEYLAST),
    dm(WM_KEYLAST),
    dm(WM_IME_STARTCOMPOSITION),
    dm(WM_IME_ENDCOMPOSITION),
    dm(WM_IME_COMPOSITION),
    dm(WM_IME_KEYLAST),
    dm(WM_INITDIALOG),
    dm(WM_COMMAND),
    dm(WM_SYSCOMMAND),
    dm(WM_TIMER),
    dm(WM_HSCROLL),
    dm(WM_VSCROLL),
    dm(WM_INITMENU),
    dm(WM_INITMENUPOPUP),

    dm(WM_IME_SETCONTEXT),
    dm(WM_IME_NOTIFY),
    dm(WM_IME_CONTROL),
    dm(WM_IME_COMPOSITIONFULL),
    dm(WM_IME_SELECT),
    dm(WM_IME_CHAR),
    dm(WM_IME_REQUEST),
    dm(WM_IME_KEYDOWN),
    dm(WM_IME_KEYUP),
    dm(WM_MOUSEHOVER),
    dm(WM_MOUSELEAVE),
    dm(WM_NCMOUSEHOVER),
    dm(WM_NCMOUSELEAVE),
    dm(WM_WTSSESSION_CHANGE),
    dm(WM_DPICHANGED),
    dm(WM_DPICHANGED_BEFOREPARENT),
    dm(WM_DPICHANGED_AFTERPARENT),
    dm(WM_GETDPISCALEDSIZE),

    dm(WM_CUT),
    dm(WM_COPY),
    dm(WM_PASTE),
    dm(WM_CLEAR),
    dm(WM_UNDO),
    dm(WM_RENDERFORMAT),
    dm(WM_RENDERALLFORMATS),
    dm(WM_DESTROYCLIPBOARD),
    dm(WM_DRAWCLIPBOARD),
    dm(WM_PAINTCLIPBOARD),
    dm(WM_VSCROLLCLIPBOARD),
    dm(WM_SIZECLIPBOARD),
    dm(WM_ASKCBFORMATNAME),
    dm(WM_CHANGECBCHAIN),
    dm(WM_HSCROLLCLIPBOARD),
    dm(WM_QUERYNEWPALETTE),
    dm(WM_PALETTEISCHANGING),
    dm(WM_PALETTECHANGED),
    dm(WM_HOTKEY),

    dm(WM_UAHDESTROYWINDOW),
    dm(WM_UAHDRAWMENU),
    dm(WM_UAHDRAWMENUITEM),
    dm(WM_UAHINITMENU),
    dm(WM_UAHMEASUREMENUITEM),
    dm(WM_UAHNCPAINTMENUPOPUP),
};
#undef dm

const char* GetWinMessageName(UINT msg) {
    for (size_t i = 0; i < dimof(gWinMessageNames); i++) {
        if (gWinMessageNames[i].msg == msg) {
            return gWinMessageNames[i].name;
        }
    }
    return "msg unknown";
}

// we might want to not show frequently posted messages
// clang-format off
uint gMsgToIgnore[] = {
    WM_NCHITTEST,
    WM_SETCURSOR,
    WM_MOUSEMOVE,
};
// clang-format on

static bool ShouldIgnoreMsg(uint msg) {
    int n = (int)dimof(gMsgToIgnore);
    for (int i = 0; i < n; i++) {
        if (gMsgToIgnore[i] == msg) {
            return true;
        }
    }
    return false;
}

void DbgLogMsg(const char* prefix, HWND hwnd, UINT msg, WPARAM wp, LPARAM lp) {
    if (ShouldIgnoreMsg(msg)) {
        return;
    }

    auto msgName = GetWinMessageName(msg);
    if (!prefix) {
        prefix = "";
    }
    logf("%shwnd: 0x%4p, msg: 0x%03x (%s), wp: 0x%x, lp: 0x%x\n", prefix, hwnd, msg, msgName, (unsigned int)wp,
         (unsigned int)lp);
}
